# (c) Copyright 2009-2012. CodeWeavers, Inc.

"""Defines classes to represent and manipulate the C4 profile information."""

import os
import re

import iso8601

import cxproduct
import cxconfig
import cxdecorators
import cxlog
import cxutils
import distversion

# for localization
from cxutils import cxgettext as _

# A mapping of the language code to the plain English name for use by the GUI.
LANGUAGES = {
    '':      _('Default'),
    'af':    _('Afrikaans'),
    'ar':    _('Arabic'),
    'as':    _('Assamese'),
    'az':    _('Azerbaijani'),
    'be':    _('Belarusian'),
    'bg':    _('Bulgarian'),
    'bn':    _('Bengali'),
    'ca':    _('Catalan'),
    'cs':    _('Czech'),
    'da':    _('Danish'),
    'de':    _('German'),
    'el':    _('Greek'),
    'en':    _('English'),
    'en-gb': _('English (Great Britain)'),
    'es':    _('Spanish'),
    'es-mx': _('Spanish (Mexico)'),
    'et':    _('Estonian'),
    'eu':    _('Basque'),
    'fa':    _('Farsi'),
    'fi':    _('Finnish'),
    'fo':    _('Faroese'),
    'fr':    _('French'),
    'gd':    _('Gaelic'),
    'gu':    _('Gujarati'),
    'he':    _('Hebrew'),
    'hi':    _('Hindi'),
    'hr':    _('Croatian'),
    'hu':    _('Hungarian'),
    'hy':    _('Armenian'),
    'id':    _('Indonesian'),
    'is':    _('Icelandic'),
    'it':    _('Italian'),
    'ja':    _('Japanese'),
    'ka':    _('Georgian'),
    'kk':    _('Kazakh'),
    'kn':    _('Kannada'),
    'ko':    _('Korean'),
    'ky':    _('Kirghiz'),
    'lt':    _('Lithuanian'),
    'lv':    _('Latvian'),
    'mk':    _('Macedonian'),
    'ml':    _('Malayalam'),
    'mn':    _('Mongolian (Cyrillic)'),
    'mr':    _('Marathi'),
    'ms':    _('Malay (Malaysia)'),
    'mt':    _('Maltese'),
    'nb':    _(u'Norwegian (Bokm\u00e5l)'),
    'ne':    _('Nepali (India)'),
    'nl':    _('Dutch'),
    'no':    _('Norwegian'),
    'or':    _('Oriya'),
    'pa':    _('Punjabi'),
    'pl':    _('Polish'),
    # pt-pt
    'pt':    _('Portuguese (Portugal)'),
    'pt-br': _('Portuguese (Brazil)'),
    'rm':    _('Romansh'),
    'ro':    _('Romanian'),
    'ru':    _('Russian'),
    'sa':    _('Sanskrit'),
    'sb':    _('Sorbian'),
    'sk':    _('Slovak'),
    'sl':    _('Slovenian'),
    'sq':    _('Albanian'),
    'sr':    _('Serbian (Latin)'),
    'st':    _('Southern Sotho'),
    'sv':    _('Swedish'),
    'sw':    _('Swahili'),
    'syr':   _('Syriac'),
    'ta':    _('Tamil'),
    'te':    _('Telugu'),
    'th':    _('Thai'),
    'tn':    _('Tswana'),
    'tr':    _('Turkish'),
    'ts':    _('Tsonga'),
    'tt':    _('Tatar'),
    'uk':    _('Ukrainian'),
    'ur':    _('Urdu'),
    'uz':    _('Uzbek'),
    'vi':    _('Vietnamese'),
    'xh':    _('Xhosa'),
    'yi':    _('Yiddish'),
    'zh':    _('Chinese'),
    'zh-cn': _('Chinese (Simplified)'),
    'zh-hk': _('Chinese (Hong Kong SAR)'),
    'zh-sg': _('Chinese (Singapore)'),
    'zh-tw': _('Chinese (Traditional)'),
    'zu':    _('Zulu'),
    }


# This is the set of environment variables which are allowed in expandable
# profile strings. The values are given for documentation only.
ENVIRONMENT_VARIABLES = (
    'accessories',     # 'Programs\Accessories'
    'allusersprofile', # 'c:\users\Public'
    'appdata',         # 'c:\users\crossover\Application Data'
    'comspec',         # 'c:\windows\system32\cmd.exe'
    'desktop',         # 'c:\users\crossover\Desktop'
    'programfiles',    # 'c:\Program Files'
    'programs',        # 'Programs'
    'startmenu',       # 'c:\users\crossover\Start Menu'
    'systemdrive',     # 'c:'
    'systemroot',      # 'c:\windows'
    'temp',            # 'c:\windows\temp'
    'userprofile',     # 'c:\users\crossover'
    'windir',          # 'c\windows'
    'winsysdir',       # 'c:\windows\system32'
    )



#####
#
# Helper functions for validating fields
#
#####

_RESTRICTED_GLOB_RE = re.compile('(?:^|/)\.[.?*]+(?:/|$)')

def _validate_file_globs(globs, location):
    """Check that the file globs don't try to escape the directory they are
    applied to.

    We do that by simply forbidding any path component that could match '..'.
    """
    # FIXME: If the string is expandable, then we could get stuff like
    #      '/.%Empty%./' that would be expanded to '/../'. It's expected that
    # the allowed environment variables won't start or end with a '.'. They
    # could end with a '/' or '\\' which could be exploited.
    for glob in globs:
        if _RESTRICTED_GLOB_RE.search(glob):
            raise AttributeError("invalid file glob '%s' in %s" % (glob, location))

def _validate_file_paths(paths, location):
    """Check that the file paths don't try to escape the directory they are
    applied to.
    """
    # FIXME: If the string is expandable, then we could get stuff like
    #      '/.%Empty%./' that would be expanded to '/../'
    for path in paths:
        # For now we simply apply the same rules as for globs since
        # '?' and '*' are not valid Windows file characters anyway.
        if _RESTRICTED_GLOB_RE.search(path):
            raise AttributeError("invalid file path '%s' in %s" % (path, location))

def _validate_regular_expression(pattern, location, allow_empty=True):
    if pattern:
        try:
            if re.match(pattern, '') and not allow_empty:
                raise AttributeError("the %s regular expression for %s is overly broad" % (pattern, location))
        except re.error:
            raise AttributeError("%s is an invalid regular expression for %s" % (pattern, location))

def _validate_language_field(field):
    # return True if there is at least one non-empty value
    for value in field.itervalues():
        if value:
            return True
    else:
        return False



#####
#
# Helper functions for dumping the profiles
#
#####

def _dump_value(out, indent, value):
    if hasattr(value, 'dump'):
        out.write(" %s\n" % type(value))
        value.dump(out, indent + "| ")
    else:
        out.write(" %s\n" % repr(value))

def _dump_fields(out, indent, obj, fields):
    for field in fields:
        if field not in obj.__dict__:
            continue
        value = obj.__dict__[field]
        if isinstance(value, dict):
            keys = value.keys()
            keys.sort()
            for key in keys:
                out.write("%s%s{%s} =" % (indent, field, key))
                _dump_value(out, indent, value[key])
        elif isinstance(value, list):
            for i in xrange(len(value)):
                out.write("%s%s[%d] =" % (indent, field, i))
                _dump_value(out, indent, value[i])
        else:
            out.write("%s%s =" % (indent, field))
            _dump_value(out, indent, value)



#####
#
# Helper functions for copying/updating C4 profile objects
#
#####

def _copy_value(value):
    try:
        copy_func = value.copy
    except AttributeError:
        if type(value) is list:
            return list(value)
        else:
            return value
    else:
        return copy_func()

def _copy_values(attributes):
    for key, value in attributes.iteritems():
        if key != 'parent':
            attributes[key] = _copy_value(value)

def _update_value(this, oth):
    try:
        update_func = this.update
    except AttributeError:
        if type(this) is list:
            this.extend(oth)
            return this
        else:
            return oth
    else:
        result = update_func(oth)
        if result is None:
            return this
        else:
            return result

def _update_values(attributes, others, exclude=()):
    for key, value in others.iteritems():
        if key in exclude:
            continue
        if key in attributes:
            attributes[key] = _update_value(attributes[key], value)
        else:
            attributes[key] = _copy_value(value)


#####
#
# C4Condition
#
#####

class C4Condition(object):
    """Represents a boolean expression."""

    @cxdecorators.abstractmethod
    def is_true(self, _properties):
        raise NotImplementedError()


class C4ConditionUnary(C4Condition):
    """A condition node with a single sub-node."""

    # A node to use to compute this condition. This field is mandatory and
    # thus defaults to None.
    child = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'child' not in self.__dict__:
            raise AttributeError("C4ConditionUnary.child is not set for application %s" % app_name)
        self.child.validate(app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('child',))


    ### Condition evaluation

    def is_true(self, properties):
        return self.child.is_true(properties)


class C4ConditionNot(C4ConditionUnary):
    """Inverts the value of the child condition node."""

    ### Condition evaluation

    def is_true(self, properties):
        return not self.child.is_true(properties)


class C4ConditionNary(C4Condition):
    """A condition node that merges the result of multiple condition subnodes.
    """

    # A list of nodes node to use to compute this condition. This field is
    # mandatory and thus defaults to [].
    children = []

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if len(self.children) == 0:
            raise AttributeError("C4ConditionNary.children is empty for application %s" % app_name)
        for child in self.children:
            child.validate(app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('children',))

    @cxdecorators.abstractmethod
    def is_true(self, properties):
        # so pylint recognizes this as an abstract class
        raise NotImplementedError()


class C4ConditionAnd(C4ConditionNary):
    """Returns True if all the children nodes evaluate to True."""

    def is_true(self, properties):
        for child in self.children:
            if not child.is_true(properties):
                return False
        return True


class C4ConditionOr(C4ConditionNary):
    """Returns False if all the children nodes evaluate to False."""

    def is_true(self, properties):
        for child in self.children:
            if child.is_true(properties):
                return True
        return False


class C4ConditionCompare(C4Condition):
    """Returns True if the specified property has the expected value."""

    OPERATORS = frozenset(('equal', 'lt', 'le', 'ge', 'gt'))

    # The property name. This field is mandatory and thus defaults to None.
    name = None

    # The property's expected value.
    # This field is mandatory and thus defaults to None.
    value = None

    def __init__(self, operator):
        C4Condition.__init__(self)
        if operator not in C4ConditionCompare.OPERATORS:
            raise AttributeError("unknown operator %s" % operator)
        self.operator = operator

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4ConditionCompare.name is not set for application %s" % app_name)
        if 'value' not in self.__dict__:
            raise AttributeError("C4ConditionCompare.value is not set for application %s" % app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('operator', 'name', 'value'))


    ### Condition evaluation

    def is_true(self, properties):
        if self.operator == 'equal':
            return properties.get(self.name, "") == self.value
        if self.operator == 'lt':
            return properties.get(self.name, "") < self.value
        if self.operator == 'le':
            return properties.get(self.name, "") <= self.value
        if self.operator == 'ge':
            return properties.get(self.name, "") >= self.value
        return properties.get(self.name, "") > self.value


class C4ConditionMatch(C4Condition):
    """Returns True if the specified property matches the given regular expression."""

    # The property name. This field is mandatory and thus defaults to None.
    name = None

    # The regular expression that the property should match.
    # This field is mandatory and thus defaults to None.
    regexp = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4ConditionMatch.name is not set for application %s" % app_name)
        if 'regexp' not in self.__dict__:
            raise AttributeError("C4ConditionMatch.regexp is not set for application %s" % app_name)

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'regexp'))


    ### Condition evaluation

    def is_true(self, properties):
        try:
            return re.match(self.regexp, properties.get(self.name, ""))
        except re.error:
            cxlog.warn("useif match pattern %s for name %s is an invalid regular expression" % (cxlog.debug_str(self.regexp), cxlog.debug_str(self.name)))
            return None



#####
#
# C4CDProfile class
#
#####

class C4CDGlob(object):
    """Contains a specific glob for detecting an application CD."""

    ### Property defaults

    # A string containing a glob to identify the files to analyze. This field
    # is mandatory and thus defaults to none.
    glob = None

    # A list of regular expressions that the selected file must all match.
    # If the list is empty, then it's enough for the file to just exist.
    patterns = tuple()


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'glob' not in self.__dict__:
            raise AttributeError("C4CDGlob.glob is not set for application %s" % app_name)
        _validate_file_globs((self.glob,), "the C4CDGlob.glob field of %s" % app_name)
        for pattern in self.patterns:
            _validate_regular_expression(pattern, "the C4CDGlob.patterns field of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('glob', 'patterns'))


class C4CDProfile(object):
    """Contains the C4 CD Profile information."""

    ### Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to none.
    parent = None

    # A list of C4CDGlobs to be used to recognize a given CD. This field is
    # mandatory and thus defaults to None.
    globs = None


    def copy(self):
        result = C4CDProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if not self.parent:
            raise AttributeError("C4CDProfile.parent is not set for application %s" % app_name)
        if not self.globs:
            raise AttributeError("C4CDProfile.globs is not set for application %s" % app_name)
        for glob in self.globs:
            glob.validate(app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('globs',))



#####
#
# C4ApplicationProfile class
#
#####


class C4RegistryGlob(object):
    """Contains a specific registry glob for detecting an installed
    application.
    """

    ### Property defaults

    # A string containing a glob to identify the registry keys to analyze.
    # This field is mandatory and thus has no default.
    # key_glob = ""

    # A string containing a regular expression to identify an optional key
    # value to analyze. If left empty, then the values are ignored. Note that
    # a key's default value is called '@'.
    value_pattern = ""

    # A string containing a regular expression to check if the value contains
    # the expected data. If left empty, then the presence of the value is
    # enough.
    data_pattern = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'key_glob' not in self.__dict__:
            raise AttributeError("C4RegistryGlob.key_glob is not set for application %s" % app_name)
        _validate_regular_expression(self.value_pattern, "the C4RegistryGlob.value_pattern field of %s" % app_name)
        if self.data_pattern != "":
            if self.value_pattern == "":
                raise AttributeError("C4RegistryGlob.data_pattern is set but not the corresponding value_pattern for application %s" % app_name)
            _validate_regular_expression(self.data_pattern, "the C4RegistryGlob.data_pattern field of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('key_glob', 'value_pattern', 'data_pattern'))


class C4ApplicationProfile(object):
    """Contains the C4 Application Profile information."""

    ### Enumerated types

    # A set of all known medals
    ALL_MEDALS = set(('gold', 'silver', 'bronze',
                      'ungold', 'unsilver', 'unbronze',
                      'knownnottowork', 'untested', 'untrusted'))

    # A list of the medals for supported applications.
    #  Note that the supported and unsupported medal list
    #  are ordered and their values correspond.  This is needed
    #  for when we need to programmatically promote or demote an app.
    SUPPORTED_MEDALS = ['gold', 'silver', 'bronze']

    # A list of the medals for supported applications.  See note
    #  about ordering above.
    UNSUPPORTED_MEDALS = ['ungold', 'unsilver', 'unbronze']

    # A set of medals known to work
    BRONZE_PLUS_MEDALS = set(('gold', 'silver', 'bronze',
                              'ungold', 'unsilver', 'unbronze'))

    UNTRUSTED_MEDAL = 'untrusted'

    ### Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to None.
    parent = None

    # A dictionary mapping the language to the application description. The
    # special '' mapping corresponds to the default application description.
    _description = {'': ''}

    # The C4 application category string. This field is mandatory and thus
    # defaults to None.
    category = None

    # A set of C4CXMedal objects represnting the times and situations this
    # application has been ranked for.
    medals = tuple()

    # A set of flag strings.
    flags = set()

    # If we know how to install this application, then this is the id of the
    # C4 entry that has the installation information.
    installer_id = None

    # If this regular expression matches an Uninstall key, then it means this
    # application is installed.
    installed_key_pattern = None

    # If this regular expression matches an Uninstall key's display name,
    # then it means this application is installed.
    installed_display_pattern = None

    # A list of C4RegistryGlobs to look for to check if the application is
    # installed.
    installed_registry_globs = tuple()

    # A list of strings containing the file globs to look for to check if the
    # application is installed.
    installed_file_globs = tuple()

    # The list of bottle types the application can be installed in.
    # The first entry is the default one.
    bottle_types = {'use': ("winxp",), 'install': ()}

    # Used to group related applications in the same bottle.
    application_group = ""

    # Maps a language to the URL where one can get the corresponding
    # installer.
    download_urls = {}

    # Maps a language to the URL of a web page from which the installer can
    # be manually downloaded.
    download_page_urls = {}

    # A list of strings containing the file globs to look for an installer file
    # that the user may have manually downloaded.
    download_globs = tuple()

    # A set of application ids that this is an upgrade or add-on for. Note
    # that this is not a dependency but a recommendation system. This can be
    # used for the following scenarios:
    # - A component which is an upgrade for another component.
    # - A recommended 'upgrade' which is a full install and thus does not
    #   depend on the item it upgrades.
    # - A codec / plugin / add-on which we recommend installing, but which
    #   technically is not an upgrade.
    extra_fors = set()

    # Optional steam id.  If this is set then an app can be installed
    # via Steam.
    steamid = None



    def copy(self):
        result = C4ApplicationProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.
        """
        if not self.parent:
            raise AttributeError("C4ApplicationProfile.parent is not set for application %s" % app_name)
        if not _validate_language_field(self._description):
            raise AttributeError("C4ApplicationProfile has no default description for application %s" % app_name)
        if not self.category:
            raise AttributeError("C4ApplicationProfile.category is not set for application %s" % app_name)
        check_template = False
        for medal in self.medals:
            medal.validate(app_name)
            if not self.parent.trusted:
                medal.medal = self.UNTRUSTED_MEDAL
            if medal.medal in self.BRONZE_PLUS_MEDALS:
                check_template = True
        if check_template and ('bottle_types' not in self.__dict__ or \
                                   not self.bottle_types['use']):
            cxlog.warn('%s has a bronze or higher medal so it should specify the bottle template(s) to use' % cxlog.to_str(app_name))
        _validate_regular_expression(self.installed_key_pattern, "the C4ApplicationProfile.installed_key_pattern field of %s" % app_name, False)
        _validate_regular_expression(self.installed_display_pattern, "the C4ApplicationProfile.installed_display_pattern field of %s" % app_name, False)
        for registryglob in self.installed_registry_globs:
            registryglob.validate(app_name)
        _validate_file_globs(self.installed_file_globs, "the installed_file_globs field of %s" % cxlog.to_str(app_name))
        _validate_file_globs(self.download_globs, "the download_globs field of %s" % cxlog.to_str(app_name))

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('_description', 'category', 'medals', 'flags',
                      'installer_id', 'installed_key_pattern',
                      'installed_display_pattern', 'installed_registry_globs',
                      'installed_file_globs', 'bottle_types',
                      'application_group', 'download_urls',
                      'download_page_urls', 'download_globs', 'extra_fors'
                      ))


    ### Accessors for special properties

    def _getdescription(self):
        """Returns the application description in the current language.

        If no description is available for the current language, then the
        default description is returned.
        """
        _lang, description = cxutils.get_language_value(self._description)
        return description

    description = property(_getdescription)


    def _getmedal(self):
        """Returns the application medal for the current platform."""
        defaultmedal = None
        product_version = cxproduct.this_product()['productversion']
        score = (99,0,0)
        for medal in self.medals:
            if medal.platform == distversion.PLATFORM :
                new_defaultmedal =  medal
                if (medal.version):
                    if (_cmp_versions(medal.version, product_version) <= 0):
                        new_score = _get_version_difference(medal.version, product_version)
                    else:
                        continue
                else:
                    new_score = (0,0,0)
                if (defaultmedal == None or new_score < score):
                    defaultmedal = new_defaultmedal
                    score = new_score

        if (defaultmedal):
            if defaultmedal.medal not in C4ApplicationProfile.SUPPORTED_MEDALS and defaultmedal.medal not in C4ApplicationProfile.UNSUPPORTED_MEDALS:
                return defaultmedal

            if not supported_list.ids():
                # No supported list, which means we don't need to mess
                #  with anything.
                return defaultmedal

            if defaultmedal.medal in C4ApplicationProfile.SUPPORTED_MEDALS and self.parent.appid not in supported_list.ids():
                # We need to de-support this.
                defaultmedal.medal = C4ApplicationProfile.UNSUPPORTED_MEDALS[C4ApplicationProfile.SUPPORTED_MEDALS.index(defaultmedal.medal)]

            if defaultmedal.medal in C4ApplicationProfile.UNSUPPORTED_MEDALS and self.parent.appid in supported_list.ids():
                # We need to support this.
                defaultmedal.medal = C4ApplicationProfile.SUPPORTED_MEDALS[C4ApplicationProfile.UNSUPPORTED_MEDALS.index(defaultmedal.medal)]

            return defaultmedal

        return None

    def _getmedal_rank(self):
        medal = self._getmedal()
        if (medal):
            return medal.medal
        else:
            return 'untested'

    medal = property(_getmedal_rank)

    def _getmedal_icon(self):
        """Returns the icon name for this application's medal."""
        return os.path.join(cxutils.CX_ROOT,
                            'share/images/medal_' + self._getmedal_rank() + '.png')

    medal_icon = property(_getmedal_icon)

    def _getmedal_date(self):
        medal = self._getmedal()
        if (medal):
            return medal.timestamp
        else:
            return ''

    medal_date = property(_getmedal_date)

    def _getis_supported(self):
        """Returns true if the application is supported."""
        try:
            return self.parent.trusted and self.medal in C4ApplicationProfile.SUPPORTED_MEDALS
        except KeyError:
            return False

    is_supported = property(_getis_supported)

    def _getis_virtual(self):
        return "virtual" in self.flags

    is_virtual = property(_getis_virtual)


    def _getpreferred_bottle_type(self):
        """Returns the profile's preferred bottle type."""
        if self.bottle_types and self.bottle_types['use'] and len(self.bottle_types['use']) > 0:
            return self.bottle_types['use'][0]
        else:
            return None

    preferred_bottle_type = property(_getpreferred_bottle_type)

    def _getis_trusted_source(self):
        return self.parent.trusted

    is_trusted_source = property(_getis_trusted_source)

    def _get_origin(self):
        return self.parent.origin

    origin = property(_get_origin)

#####
#
# C4InstallerProfile class
#
#####

class C4EnvVar(object):
    """Describes an environment variable to set."""

    ### Property defaults

    # The name of the environment variable to set.
    # This field is mandatory and thus has no default.
    # name = ""

    # The value of the environment variable to set.
    # This field is mandatory and thus has no default.
    # value = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4Environment.name is not set for application %s" % app_name)
        if 'value' not in self.__dict__:
            raise AttributeError("C4Environment.value is not set for application %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'value'))


class C4FilesToCopy(object):
    """Specifies a set of files to copy."""

    ### Property defaults

    # A glob identifying the files to copy.
    # This field is mandatory and thus defaults to None.
    glob = None

    # A restricted path specifying where to copy the above files.
    # This field is mandatory and thus defaults to None.
    destination = None


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'glob' not in self.__dict__:
            raise AttributeError("C4FilesToCopy.glob is not set for application %s" % app_name)
        _validate_file_globs((self.glob,), "the C4FilesToCopy.glob field of %s" % app_name)
        if 'destination' not in self.__dict__:
            raise AttributeError("C4FilesToCopy.destination is not set for application %s" % app_name)
        _validate_file_paths((self.destination,), "the C4FilesToCopy.destination field of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('glob', 'destination'))


class C4RegistryValue(object):
    """Specifies a registry value to set."""

    ### Property defaults

    # The name of the registry value to set.
    # This field is mandatory and thus has no default.
    # name = ""

    # The content of the registry value to set.
    # This field is mandatory and thus has no default.
    # data = ""


    ### Instance validation and dumping

    def validate(self, app_name, key):
        """Checks that all the mandatory fields have been set."""
        if 'name' not in self.__dict__:
            raise AttributeError("C4RegistryValue.name is not set for %s in application %s" % (key, app_name))
        if 'data' not in self.__dict__:
            raise AttributeError("C4RegistryValue.data is not set for %s in application %s" % (key, app_name))
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'data'))


class C4Registry(object):
    """Specifies a set of registry values to set."""

    ### Property defaults

    # The name of the registry key.
    # This field is mandatory and thus defaults to None.
    key = None

    # The name of the registry value to set if any.
    values = tuple()


    ### Instance validation and dumping

    ROOT_KEYS = frozenset((u'HKEY_CLASSES_ROOT', u'HKCR',
                           u'HKEY_CURRENT_USER', u'HKCU',
                           u'HKEY_LOCAL_MACHINE', u'HKLM',
                           u'HKEY_USERS', u'HKU'))

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if not self.key:
            raise AttributeError("C4Registry.key is not set for application %s" % app_name)
        if u'\\' not in self.key:
            raise AttributeError("C4Registry.key %s contains no backslashes for application %s" % (self.key, app_name))
        if self.key.split(u'\\', 1)[0] not in self.ROOT_KEYS:
            raise AttributeError("C4Registry.key %s has an invalid root key for application %s" % (self.key, app_name))
        if self.values:
            for value in self.values:
                value.validate(app_name, self.key)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('key', 'values'))


class C4LnkFile(object):
    """Specifies a wnidows shortcut to create."""

    ### Property defaults

    # The path of the shortcut file to create.
    # This field is mandatory and thus defaults to None.
    shortcut = None

    # The path of the file the shortcut should point to.
    # This field is mandatory and thus defaults to None.
    target = None

    # The work directory to store in the shortcut if any.
    workdir = ""


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set."""
        if 'shortcut' not in self.__dict__:
            raise AttributeError("C4LnkFile.shortcut is not set for application %s" % app_name)
        if 'target' not in self.__dict__:
            raise AttributeError("C4LnkFile.target is not set for %s in application %s" % (self.shortcut, app_name))
        _validate_file_paths((self.shortcut, self.target), "the C4LnkFile shortcut or target fields of %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('shortcut', 'target', 'workdir'))


class C4InstallerProfile(object):
    """Contains the C4 Installer Profile information."""

    ### 'Pre-Install' Property defaults

    # Points to the umbrella C4Profile object that ties a C4 entry's
    # specialized profiles together. This field is mandatory and thus defaults
    # to None.
    parent = None

    # A string storing the application id of the profile these settings apply
    # to. Note that it may not match the application id of the parent
    # C4Profile. It is optional and defaults to None.
    appid = None

    # Specifies when this installer profile should be used.
    use_if = None

    # Used to warn users about issues they may encounter during the
    # installation.
    _installation_notes = None

    # A file that should be present on the DVD. If it is missing, then the UDF
    # side has been mounted and it needs remounting.
    udf_remount_check = ""

    # List of names under which the installer may be found on local media.
    local_installer_file_globs = tuple()

    # The list of applications that must be installed before we can install
    # this one.
    pre_dependencies = frozenset()

    # The list of applications that must be installed after installed this one
    # in order for it to actually work.
    post_dependencies = frozenset()

    # The important cxdiag warnings, or those to ignore.
    cxdiag_checks = tuple()

    ### 'Install' Property defaults

    # The file globs to use to find the installer when faced with a directory.
    installer_file_globs = tuple()

    # A full basename to treat the installer as despite what its real name
    # may be
    installer_treatas = ""

    # A list of options to pass to the application installer.
    installer_options = tuple()

    # A list of options to pass to the application installer when doing a
    # silent installation.
    installer_silent_options = tuple()

    # The windows version to use to run the installer.
    installer_winver = ""

    # A Wine DllOverrides string to use to run the installer.
    installer_dlloverrides = ""

    # A list of C4EnvVar objects describing the environment variables to set
    # before running the installer.
    installer_environment = tuple()

    # A size in KB above which the installer is considered to be a
    # self-extracting archive.
    selfextract_threshold = None

    # A list of options to pass to the self-extracting archive to get it to
    # extract in a known location. These options can make use of the special
    # %ExtractDir% environment variable.
    selfextract_options = tuple()

    # A list of options to pass to the self-extracting archive when doing a
    # silent installation.
    selfextract_silent_options = tuple()

    # Specifies the encoding of the file names stored in the zip archive.
    # Blank by default which means no conversion is done.
    zip_encoding = ""

    # A list of C4FilesToCopy objects describing how to copy the files for
    # applications that don't have a proper installer.
    files_to_copy = tuple()

    # If True, force a reboot after the installer completes.
    post_install_reboot = False

    # A list of C4Registry objects describing the registry keys and values to
    # create before running the installer.
    pre_install_registry = tuple()

    # A list of fake dlls to remove before running the installer.
    pre_rm_fake_dlls = tuple()

    # A list of Netscape plugin dlls to export before running the installer.
    pre_install_nsplugin_dlls = tuple()

    # A list of Netscape plugins to ignore.
    ignore_nsplugin_dlls = tuple()

    # A list of C4LnkFile objects describing the Windows shortcuts to be
    # created.
    lnk_files = tuple()

    ### 'Post-Install' Property defaults

    # Set to True if the application is known to not create any menu.
    skip_menu_creation = False

    # Set to True if the application is known to not create any association.
    skip_assoc_creation = False

    # Set to True if the installer should export browser plugins to the host.
    install_nsplugins = False

    # The list of EAssocs that should be made the default.
    default_eassocs = tuple()

    # The list of EAssocs that should be exported as alternatives.
    alt_eassocs = tuple()

    # A list of dll file paths to register after install
    post_registerdll = tuple()

    # A web page to open after the application has installed.
    post_install_urls = None


    def __init__(self):
        self._installation_notes = {}
        self.pre_dependencies = set()
        self.post_dependencies = set()
        self.post_install_urls = {}

    def copy(self):
        result = C4InstallerProfile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__, exclude=('parent',))


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.
        """
        if not self.parent:
            raise AttributeError("C4InstallerProfile.parent is not set for application %s" % app_name)
        _validate_file_globs(self.local_installer_file_globs, "the local_installer_file_globs field of %s" % app_name)
        for dependency in self.pre_dependencies:
            if dependency.startswith('.'):
                raise AttributeError("C4InstallerProfile.pre_dependencies '%s' must not start with a dot for application %s" % (dependency, app_name))
        for dependency in self.post_dependencies:
            if dependency.startswith('.'):
                raise AttributeError("C4InstallerProfile.prost_dependencies '%s' must not start with a dot for application %s" % (dependency, app_name))
        _validate_file_globs(self.installer_file_globs, "the installer_file_globs field of %s" % app_name)
        for envvar in self.installer_environment:
            envvar.validate(app_name)
        if self.selfextract_threshold is None:
            if self.selfextract_options:
                raise AttributeError("C4InstallerProfile.selfextract_options is set but no threshold has been specified for application %s" % app_name)
            if self.selfextract_silent_options:
                raise AttributeError("C4InstallerProfile.selfextract_silent_options is set but no threshold has been specified for application %s" % app_name)
        for file_copy in self.files_to_copy:
            file_copy.validate(app_name)
        for regentry in self.pre_install_registry:
            regentry.validate(app_name)
        if self.lnk_files:
            for lnk_file in self.lnk_files:
                lnk_file.validate(app_name)
            if self.skip_menu_creation:
                raise AttributeError("C4InstallerProfile.lnk_files must be empty when skip_menu_creation is True for application %s" % app_name)
        if self.skip_assoc_creation and (self.default_eassocs or self.alt_eassocs):
            raise AttributeError("C4InstallerProfile.default_eassocs and alt_eassocs must be empty when skip_assoc_creation is True for application %s" % app_name)
        for eassocs in (self.default_eassocs, self.alt_eassocs):
            for eassocid in eassocs:
                if eassocid[0] != '.' or '\\' in eassocid:
                    raise AttributeError("C4InstallerProfile.default_eassocs and alt_eassocs must must all start with a dot and not contain backslashes (%s) for application %s" % (eassocid, app_name))

        return True


    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('use_if', 'installation_notes', 'udf_remount_check',
                      'local_installer_file_globs', 'pre_dependencies',
                      'post_dependencies',
                      'cxdiag_checks', 'installer_file_globs',
                      'installer_treatas', 'installer_options',
                      'installer_silent_options', 'installer_winver',
                      'installer_dlloverrides', 'installer_environment',
                      'selfextract_threshold', 'selfextract_options',
                      'selfextract_silent_options', 'zip_encoding',
                      'files_to_copy', 'post_install_reboot',
                      'pre_install_registry', 'pre_rm_fake_dlls',
                      'pre_install_nsplugin_dlls', 'ignore_nsplugin_dlls',
                      'lnk_files', 'skip_menu_creation', 'skip_assoc_creation',
                      'install_nsplugins', 'default_eassocs', 'alt_eassocs',
                      'post_registerdll', 'post_install_urls'))


    ### Accessors for special properties

    def use(self, use_if_props):
        if self.appid is not None and self.appid != use_if_props['appid']:
            return False
        if self.use_if is None:
            return True
        return self.use_if.is_true(use_if_props)


    def _getinstallation_notes(self):
        """Returns the application installation notes in the current language.

        If no installation notes are available for the current language, then
        the default installation notes are returned.
        """
        _lang, installation_notes = cxutils.get_language_value(self._installation_notes)
        return installation_notes

    installation_notes = property(_getinstallation_notes)


    def get_default_eassocs_str(self):
        """Returns a string concatenating all the EAssocs that are to be made
        defaults.

        The string is suitable for cxassoc.
        """
        return ":".join(self.default_eassocs)


    def get_alt_eassocs_str(self):
        """Returns a string concatenating all the EAssocs that are to be made
        alternatives.

        The string is suitable for cxassoc.
        """
        return ":".join(self.alt_eassocs)


    def _getpost_install_url(self):
        """Returns the postinstall web page to open for the current language.

        If no URL is specified for the current language, then the default
        URL is returned.
        """
        _lang, post_install_url = cxutils.get_language_value(self.post_install_urls)
        return post_install_url

    post_install_url = property(_getpost_install_url)


#####
#
# Umbrella C4Profile class
#
#####

class C4CXVersion(object):
    """Identifies a product plus version combination which is compatible with
    the corresponding profile.
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be immutable objects,

    # A string storing the product name.
    # This field is mandatory and thus defaults to None.
    product = None

    # A string storing the version prefix.
    # This field is mandatory and thus defaults to None.
    cxversion = None


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.product:
            raise AttributeError("C4Profile.product is not set for application %s" % app_name)
        if not self.cxversion:
            raise AttributeError("C4Profile.cxversion is not set for application %s" % app_name)
        # FIXME: We should check that the cxversion field is valid
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('product', 'cxversion'))

    def copy(self):
        result = C4CXVersion()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)


class C4PTimestamp(object):
    datetime = None
    
    def __init__(self, timestamp):
        if timestamp is not None:
            self.datetime = iso8601.parse_date(timestamp)

    def __cmp__(self, oth):
        # We want "better" timestamps to be less.
        
        if not isinstance(oth, C4PTimestamp):
            raise TypeError()

        # If one object has a timestamp, prefer that one.
        if self.datetime is None:
            if oth.datetime is None:
                return 0
            else:
                return 1

        if oth.datetime is None:
            return -1

        # Else, we want later timestamps to be lesser, which is the opposite of
        # how datetime objects compare.
        return cmp(oth.datetime, self.datetime)

class C4CXContributor(object):
    """Identifies a contributor to the profile
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be immutable objects,

    # A string storing the contributor name.
    # This field is mandatory and thus defaults to None.
    name = None

    # A string storing the ID of the contributor in our database
    userid = 0


    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.name:
            raise AttributeError("C4Profile.name is not set for application %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('name', 'userid'))

    def copy(self):
        result = C4CXContributor()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)

class C4CXMedal(object):
    """Identifies a contributor to the profile
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be immutable objects,

    # A string storing the rank, Must be in all medals set.
    # This field is mandatory and thus defaults to None.
    medal = None

    # A string storing the platform this medal is for.
    platform = None

    # A the version this medal was ranked for
    version = None

    # the number of rankings this medal represents
    number = 0

    # the timestamp of the last time this medal was ranked
    timestamp = None

    ### Instance validation and dumping

    def validate(self, app_name):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.medal:
            raise AttributeError("C4Profile.medal is not set for application %s" % app_name)
        if self.medal not in C4ApplicationProfile.ALL_MEDALS:
            raise AttributeError("C4Profile.medal invalid medal for application %s" % app_name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('medal', 'platform', 'version', 'number', 'timestamp'))

    def copy(self):
        result = C4CXMedal()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)

### Helper for computing the profile scores

def _get_version_difference(a, b):
    a = a.split('.')
    b = b.split('.')
    if len(a) < len(b):
        a.extend([0] * (len(b) - len(a)))
    elif len(b) < len(a):
        b.extend([0] * (len(a) - len(b)))
    
    result = []
    sign = 0
    for i in range(len(a)):
        if sign == 0:
            sign = cmp(int(a[i]), int(b[i]))
        result.append((int(a[i]) - int(b[i])) * sign)
    return tuple(result)

def _cmp_versions(a, b):
    a = a.split('.')
    b = b.split('.')
    if len(a) < len(b):
        a.extend([0] * (len(b) - len(a)))
    elif len(b) < len(a):
        b.extend([0] * (len(a) - len(b)))
    
    for i in range(len(a)):
        result = cmp(int(a[i]), int(b[i]))
        if result != 0:
            return result
    return 0


builtin, download, dropin = range(3)
class C4Profile(object):
    """This is an umbrella class which ties a C4 entry's CD, Application and
    Installer profiles together.

    It also stores the common parts between these, that is the application id
    and its name.
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be immutable objects,

    # this field does not appear in the C4P file. It reflects if this profile
    # has been parsed from a trusted source or not. It is set by the
    # c4parser
    trusted = False

    # this field does not appear in the C4P file. It reflects if this profile
    # originated from the .c4ps that were bundled with CrossOver or if
    # they were added at runtime via an association or drag-n-drop.
    source = builtin

    # at present, this field does not appear in the C4P file.
    # A string representing the origin of the c4profile.
    # It is set by the c4parser
    origin = None

    # A string storing the application id, usually a C4 identifier.
    # This field is mandatory and thus defaults to None.
    appid = None

    # The timestamp of the profile's last modification.
    # This field is optional and defaults to None.
    timestamp = None

    # The CrossOver products and versions this profile was tested with.
    cxversions = tuple()

    # A dictionary mapping the language to the application name in that
    # language. The special '' mapping corresponds to the default application
    # name. This field is mandatory and thus defaults to None.
    localized_names = None

    # a set of contributors to the tie file
    contributor = tuple()

    # Points to the CD profile information if any.
    # In pysetup we may never use this field but it would be used by the CD
    # detection tool.
    cd_profile = None

    # Points to the application profile information if any.
    app_profile = None

    # The list of installer profiles. There may be more than one, because some
    # may be modifying properties of other profiles, and some may only be used
    # in specific conditions.
    installer_profiles = tuple()


    ### Instance validation and dumping

    def validate(self):
        """Checks that all the mandatory fields have been set on this profile
        and in the connected profiles.

        Note that this only checks internal consistency, not consistency with
        other profiles.
        """
        if not self.appid:
            raise AttributeError("C4Profile.appid is not set")
        # pylint: disable-msg=E1101
        elif self.appid.startswith('.'):
            raise AttributeError("C4Profile.appid '%s' must not start with a dot" % self.appid)
        if not _validate_language_field(self.localized_names):
            raise AttributeError("C4Profile.localized_names is not set for %s" % self.appid)
        for cxversion in self.cxversions:
            cxversion.validate(self.name)

        # Pylint does not see that if these are not None, then then will point
        # to instances that do have the validate function()
        # pylint: disable-msg=E1101
        if self.cd_profile:
            self.cd_profile.validate(self.name)
        if self.app_profile:
            self.app_profile.validate(self.name)
        for inst_profile in self.installer_profiles:
            inst_profile.validate(self.name)
        return True

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self,
                     ('appid', 'timestamp', 'cxversions', '_name', 'app_profile',
                      'installer_profiles', 'cd_profile'))


    def copy(self):
        result = C4Profile()
        result.__dict__.update(self.__dict__)
        _copy_values(result.__dict__)
        if result.cd_profile:
            result.cd_profile.parent = result
        if result.app_profile:
            result.app_profile.parent = result
        for inst_profile in result.installer_profiles:
            inst_profile.parent = result

        return result

    def update(self, oth):
        _update_values(self.__dict__, oth.__dict__)
        # The profiles may have been copied, in which case the parent needs to
        # be updated.
        if self.cd_profile:
            self.cd_profile.parent = self
        if self.app_profile:
            self.app_profile.parent = self
        for inst_profile in self.installer_profiles:
            inst_profile.parent = self


    ### Helpers to compute the profile score

    # The product's code and numerical version data
    _product = None
    _product_version = None
    _product_missing = None

    # A cache mapping the (product, cxversion) tuples to the corresponding
    # score
    _scores = {}


    ### Accessors for special properties

    def _getname(self):
        """Returns the application name in the current language.

        If no name is available for the current language, then the
        default name is returned.
        """
        _lang, name = cxutils.get_language_value(self.localized_names)
        return name

    name = property(_getname)

    _score = None


    def _version_score_for_product(self, version, product):
        if version.product == product['productid']:
            product_mismatch = 0
        else:
            product_mismatch = 1

        # This case is a bit of a hack.  Since as of CrossOver 11 the cxoffice product
        #  has swallowed up the games product, we need to match cxgames crossties as well.
        if product['productid'] == 'cxoffice' and version.product == 'cxgames':
            product_mismatch = 0

        if self.source == dropin:
            source = 0
        else:
            source = 1
        
        if _cmp_versions(version.cxversion, product['productversion']) > 0:
            old_product = 1
        else:
            old_product = 0
        
        version_difference = _get_version_difference(version.cxversion, product['productversion'])
        
        timestamp = C4PTimestamp(self.timestamp)
        
        return (product_mismatch, source, old_product, version_difference, timestamp)

    def score_for_product(self, product):
        """A score that defines the profile's adequation to the given
        product. The lower the score the better.
        
        Product is a dictionary containing keys named productid and version.
        
        The result is actually a tuple of
        (product_mismatch, source, old_product, version_difference, timestamp)
        
        Where
        * product_mismatch is 0 if the profile matches the current product and 1
          if it does not.
        * old_product is 1 if the product is too old for this profile
        * source is 0 if the profile is a dropin, 1 if it's from a builtin tie
        * version_difference is a tuple of the magnitude of differences between
          components of the version numbers
        * timestamp is a C4PTimestamp object representing the timestamp"""
        return min(self._version_score_for_product(cxversion, product) for cxversion in self.cxversions)

    def _get_score(self):
        """A score that defines the profile's adequation to the current
        product. See score_for_product for the score's meaning."""
        if self._score is None:
            self._score = self.score_for_product(cxproduct.this_product())
        return self._score

    score = property(_get_score)

    def _is_for_current_product(self):
        """Returns True if one of the cxversion properties match the current
        product (though not necessarily this version), and False otherwise.
        """
        if self.score[0] == 1:
            return False
        return True

    is_for_current_product = property(_is_for_current_product)

    def usable_with_product(self, product):
        score = self.score_for_product(product)
        return score[0] == 0 and score[2] == 0


    def _getis_ranked(self):
        if self.app_profile and \
            self.app_profile.medal not in ('knownnottowork', 'untested'):
            return True
        if self.source == dropin:
            return True
        return False

    is_ranked = property(_getis_ranked)


    def _getis_revoked(self):
        revokelist_file = get_revokelist_file()
        if self.appid.encode('utf8') in revokelist_file:
            for timestamp_range in revokelist_file[self.appid.encode('utf8')].values():
                start_timestamp, end_timestamp = timestamp_range.split('..', 1)
                if self.timestamp is None:
                    if start_timestamp == '':
                        return True
                else:
                    timestamp = iso8601.parse_date(self.timestamp)
                    if (start_timestamp == '' or iso8601.parse_date(start_timestamp) <= timestamp) and \
                       (end_timestamp == '' or iso8601.parse_date(end_timestamp) >= timestamp):
                        return True
        return False

    is_revoked = property(_getis_revoked)


    def _getdetails_url(self):
        """Returns the url to a page with more information about this profile,
        or '' if there is no such page."""
        # For now, just treat com.codeweavers.c4.NNN as special.
        # In the future, we may want to add a field in c4p files for this.
        # pylint: disable-msg=E1101
        if self.appid.startswith('com.codeweavers.c4.'):
            c4_id = self.appid[19:]
            if c4_id.isdigit():
                return 'http://www.codeweavers.com/compatibility/browse/name?app_id=%s' % c4_id
        return ''

    details_url = property(_getdetails_url)


    def _is_application(self):
        """Returns True if the application flag is set."""
        # pylint: disable-msg=E1101
        return 'application' in self.app_profile.flags

    is_application = property(_is_application)


    def _is_component(self):
        """Returns True if the component flag is set."""
        # pylint: disable-msg=E1101
        return 'component' in self.app_profile.flags

    is_component = property(_is_component)


    def _is_nsplugin(self):
        """Returns True if the nsplugin flag is set."""
        # pylint: disable-msg=E1101
        return 'nsplugin' in self.app_profile.flags

    is_nsplugin = property(_is_nsplugin)

    def _get_contributors(self):
        contributors = ""
        for contributor in self.contributor:
            if (contributors):
                contributors += ", " + contributor.name
            else:
                contributors = contributor.name
        return contributors.encode('utf8')

    contributors = property(_get_contributors)

    def _get_contributors_html(self):
        contributors = ""
        for contributor in self.contributor:
            if (contributors):
                if (contributor.userid):
                    contributors += ", " + '<a href="http://www.codeweavers.com/profile/?id='+contributor.userid+'">' + cxutils.html_escape(contributor.name.encode('utf8')) + "</a>"
                else:
                    contributors += ", " + cxutils.html_escape(contributor.name.encode('utf8'))
            else:
                if (contributor.userid):
                    contributors = '<a href="http://www.codeweavers.com/profile/?id='+contributor.userid+'">' + cxutils.html_escape(contributor.name.encode('utf8')) + "</a>"
                else:
                    contributors = cxutils.html_escape(contributor.name.encode('utf8'))
        return contributors.encode('utf8')

    contributors = property(_get_contributors)
    contributors_html = property(_get_contributors_html)


#####
#
# C4RevokeProfile class
#
#####
class C4RevokeProfile(object):
    """Contains the identifying information of the C4Profile instances to
    revoke.
    """

    ### Property defaults

    # These class variables provide defaults in case the corresponding
    # instance variable is not set. They should all be immutable objects,

    # A string storing the application id, usually a C4 identifier.
    # This field is mandatory and thus defaults to None.
    appid = None

    # The timestamp of the most oldest revision of the profile to revoke.
    # This field is optional and defaults to None.
    first = None

    # The timestamp of the most recent revision of the profile to revoke.
    # This field is optional and defaults to None.
    last = None


    ### Instance validation and dumping

    def validate(self):
        """Checks that all the mandatory fields have been set on this profile.
        """
        if not self.appid:
            raise AttributeError("C4RevokeProfile.appid is not set")

    def dump(self, out, indent=""):
        _dump_fields(out, indent, self, ('appid', 'first', 'last'))


# supported_list is implemented as a singleton -- it
#  will read itself on first access.
#
# supported_list is a simple list of app ids.  If supported_list
#  is non-NULL, then /only/ apps in the list should appear
#  with a 'supported' medal in the GUI.
class supported_list(object):
    _supported_list_read = False
    _supported_list = None

    @classmethod
    def ids(cls):
        if not cls._supported_list_read:
            cls.parse_supported_list()
            cls._supported_list_read = True
        return cls._supported_list

    @classmethod
    def parse_supported_list(cls):
        filename = os.path.join(cxutils.CX_ROOT, "share/crossover/data", ".supported")

        try:
            f = open(filename, 'U')
        except IOError:
            # No such file
            return

        try:
            cls._supported_list = []
            for line in f:
                line = line.strip()
                if line != '' and not line.startswith('#'):
                    cls._supported_list.append(line.decode('utf8'))
        finally:
            f.close()


_REVOKELIST_FILE = None

def get_revokelist_file():
    """Returns a cxconfig file used to store all the revokelist entries we've seen."""
    # pylint: disable-msg=W0603
    global _REVOKELIST_FILE
    if _REVOKELIST_FILE is None:
        _REVOKELIST_FILE = cxconfig.File(os.path.join(cxproduct.get_user_dir(), 'tie', 'crossover.revokelist'))
    return _REVOKELIST_FILE
