# (c) Copyright 2009, 2012. CodeWeavers, Inc.

"""Handles refreshing and reading the C4 profile information."""

import os
import threading
import time
import UserDict

import cxobjc
import cxlog
import cxutils
import cxurlget
import cxproduct
import datetime
import subprocess
import tempfile
import binascii
import c4profiles

import c4parser
import shutil


UNKNOWN = "com.codeweavers.unknown"
# This is the application id of the unknown application profile


class _C4PSource(object):
    """This is a helper class for C4ProfilesSet.

    It keeps track of a couple of extra data about a C4Pfile object: its old
    list of profiles and its last modification time.

    The 'trusted' flag marks profiles that are official
    CodeWeavers profiles.  Untrusted profiles will produce a 
    user warning upon selection.  Ideally this will be handled via 
    signing but at the moment it is based on where the files came from.

    The 'source' member marks profiles which were added at runtime via
    associations or drag and drop.  It's used for ranking when
    there are profile collisions.  Possible values are 
    builtin, download, dropin
    """

    def __init__(self, c4pfile, trusted=True, source=c4profiles.builtin):
        self.c4pfile = c4pfile
        self.profiles = {}
        self.mtime = 0
        self.trusted = trusted
        self.source = source


class C4Profiles(cxobjc.Proxy):
    pass

@cxobjc.method(C4Profiles, 'profilesSet')
def profilesSet():
    return C4ProfilesSet.all_profiles()

@cxobjc.method(C4Profiles, 'dropinC4PFile_withSig_')
def dropin_C4PFile(c4pfilename, sigfilepath):
    if sigfilepath:
        sigfilepath = sigfilepath.encode('utf8')
    return C4ProfilesSet.add_c4p_file(c4pfilename.encode('utf8'), sigfilepath, c4profiles.dropin)

@cxobjc.method(C4Profiles, 'tasteC4PFile_')
def taste_c4p(c4pfilename):
    return C4ProfilesSet.taste_new_c4p_file(c4pfilename)

@cxobjc.method(C4Profiles, 'prepareC4PFile_')
def prepare_c4p(c4pfilename):
    return C4ProfilesSet.prepare_external_c4p(c4pfilename)

@cxobjc.method(C4Profiles, 'prepareSingleUseC4PFile_')
def prepare_temp_c4p(c4pfilename):
    return C4ProfilesSet.prepare_external_c4p(c4pfilename, True)

@cxobjc.method(C4Profiles, 'updateOnlineProfilesFromURL_')
def updateOnlineProfiles(url):
    return (update_online_profiles(url))

class C4ProfilesSet(UserDict.IterableUserDict):
    """A mapping of application ids to the corresponding C4Profile objects."""

    #####
    #
    # Instance methods
    #
    #####

    def __init__(self):
        """Adds a data mapping attribute that third parties can use to store
        or cache data related to this profiles set.
        """
        UserDict.IterableUserDict.__init__(self)
        self.userdata = {}


    def installer_profiles(self, appid):
        """Returns the application's installer profiles if any, and the
        unknown application installer profiles otherwise.
        """
        installers = None
        if appid in self:
            profile = self[appid]
            if profile.app_profile and profile.app_profile.installer_id:
                try:
                    profile = self[profile.app_profile.installer_id]
                except KeyError:
                    profile = None
            if profile:
                installers = profile.installer_profiles
        if not installers:
            installers = self[UNKNOWN].installer_profiles
        return installers

    installers_ = installer_profiles

    def unknown_installer(self):
        return self[UNKNOWN].installer_profiles[0]

    def unknown_profile(self):
        return self[UNKNOWN]

    def contentsDict(self):
        return self.data


    #####
    #
    # Class methods and data
    #
    #####

    _lock = threading.Lock()
    # To avoid races while updating _sources.

    _sources = {}
    # Keep track of the files the profiles come from. This is so we know when
    # to refresh the profiles set, and also so we know to delete a given
    # profile if the file goes away. This is a mapping so each file is present
    # only once (and for now order does not matter).

    _last_checked = 0
    # The timestamp of the last time the freshness of the _profiles data
    # was checked

    _profiles = None
    # The cached profiles set. This variable is updated atomically, so this
    # set can be returned at any time without locking.

    _revoked_licenses = set()
    # The set of revoked licenses. This is intentionally static (no need to
    # track it separately for each set) and is updated whenever profiles are
    # refreshed.

    _mtimes = {}
    # We keep track of the mtimes of each file in order to
    # update our list when needed.  The keys are source files.

    tiedir = os.path.join(cxproduct.get_user_dir(), "tie")
    _tie_files_loaded = False

    @classmethod
    def add_c4p_file(cls, c4pfilename, sigfile=None, source=c4profiles.builtin):
        """Reads the specified c4p file and add its profiles to the global
        database.

        The c4p file is then watched for updates and re-read if it is modified.
        """
        trusted = cls.is_c4p_file_trusted(c4pfilename, sigfile)
        c4pfile = c4parser.C4PFile(c4pfilename, trusted, source)
        cls._sources[c4pfile.filename] = _C4PSource(c4pfile, trusted, source)
        cls._last_checked = 0
        return c4pfile

    @classmethod
    def is_c4p_file_trusted(cls, filepath, sigfilepath=None):
        """ Evaluate the trustworthiness of a c4p.  Typically the
        signature filename (if it exists) can be derived from
        the filepath.  When using temp files (e.g. via taste_new_c4p_file)
        the sigfile may have been uniquely generated in which case
        it needs to be specified here.
        """

        if not filepath:
            return False

        # Files that originate from within the CrossOver package
        #  get a free pass.
        if filepath.startswith(cxutils.CX_ROOT):
            return True

        if not sigfilepath:
            sigfilepath = filepath + ".sig"

        if not os.path.exists(sigfilepath):
            return False

        keyfilepath = os.path.join(cxutils.CX_ROOT, "share/crossover/data", "tie.pub")

        args = ["openssl", "dgst", "-sha1", "-verify", keyfilepath, "-signature", sigfilepath, filepath]
        verification_result = cxutils.run(args, stdout=cxutils.NULL, stderr=cxutils.NULL)
        if verification_result[0] == 0:
            return True
        else:
            cxlog.err("The file %s is signed, but the signature is invalid." % filepath)
            return False

    @classmethod
    def prepare_external_c4p(cls, filename, temp=False):
        """Given a downloaded or dropped-in c4p, get it ready for loading.
        This might involve unzipping and/or stripping out a signature.

        Returns the new [filepath, sigpath].  If temp = False
        then sigpath will always be filepath + ".sig".

        In the event of catastrophe, filepath will be None.
        """

        c4pfilebase = os.path.splitext(os.path.basename(filename))[0]
        gunzip_proc = None
        if os.path.splitext(filename)[1] == ".gz":
            c4pfilebase = os.path.splitext(c4pfilebase)[0]
            # -d = --decompress
            # -c = --to-stdout
            # Using single-letter flags for purposes of backwards-compatibility.
            args = ["gzip", "-d", "-c", filename]
            gunzip_proc = subprocess.Popen(args, stdout=subprocess.PIPE)
            openfile = gunzip_proc.stdout
        else:
            openfile = open(filename, "r")

        try:
            os.makedirs(cls.tiedir)
        except os.error:
            pass

        [opendatafd, datafilepath] = tempfile.mkstemp(suffix=".tie", prefix=c4pfilebase)
        opendata = os.fdopen(opendatafd, "w+b")

        sigannounce = "<!-- Signature"
        signed = False
        sigfilepath = None
        for l in openfile:
            if sigannounce in l:
                [opensigfd, sigfilepath] = tempfile.mkstemp(suffix=".sig", prefix=datafilepath)
                opensig = os.fdopen(opensigfd, "w+b")
                signed = True

            if signed:
                # We have a signature embedded in an XML comment.
                # Strip out the commenty bits.
                strippedline = l.replace(sigannounce, "")
                strippedline = strippedline.replace("-->", "")
                strippedline = strippedline.rstrip()
                splitline = strippedline.split("= ")
                if len(splitline) > 1:
                    hexline = splitline[1].strip()
                else:
                    hexline = strippedline.strip()

                opensig.write(binascii.unhexlify(hexline))
            else:
                opendata.write(l) 

        if gunzip_proc:
            gunzip_proc.communicate()
            if gunzip_proc.returncode != 0:
                cxlog.err("an error occurred while unzipping %s. This file will be ignored." % filename)
                return [None, None]
        else:
            openfile.close()

        if signed:
            opensig.close()

        opendata.close()

        if not temp:
            finaldatafilepath = os.path.join(cls.tiedir, c4pfilebase + ".tie")
            finalsigfilepath = os.path.join(cls.tiedir, c4pfilebase + ".tie.sig")

            # The order of this is a bit important.  If we are interrupted
            # between renames, it's better to have a stray (meaningless) sig
            # than a datafile without a sig.
            if sigfilepath:
                shutil.move(sigfilepath, finalsigfilepath)
            shutil.move(datafilepath, finaldatafilepath)

        else:
            finaldatafilepath = datafilepath
            finalsigfilepath = sigfilepath

        return [finaldatafilepath, finalsigfilepath]

    @classmethod
    def taste_new_c4p_file(cls, filepath):
        """For GUI purposes it is useful to be able to detect if a file will be
        trusted or untrusted before we actually load it into any CrossOver
        dirs. This function prepares and tests the signature of a c4p file
        using disposable temp files.

        Returns True if the file is trusted.
        """
        [c4p, sig] = cls.prepare_external_c4p(filepath, temp=True)
        trusted = cls.is_c4p_file_trusted(c4p, sig)
        if c4p:
            os.remove(c4p)
        if sig:
            os.remove(sig)

        return trusted


    @classmethod
    def all_profiles(cls):
        """Returns a mapping of application ids to the corresponding C4Profile
        objects.

        This automatically reads or re-reads the relevant c4p files if needed
        but will always return a consistent view of the underlying c4p profile
        database.
        """
        if not cls._tie_files_loaded:
            cls._lock.acquire()
            if not cls._tie_files_loaded:
                cls.load_tie_files()
                cls._tie_files_loaded = True
            cls._lock.release()

        # The consistency guarantee means that even if we split the local C4
        # database across many different c4p files, and these are in various
        # stages of being updated, the returned profiles will not mix old data
        # with new data.
        if cls._profiles is None or cls._last_checked + 2 < time.time():
            cls._lock.acquire()
            try:
                # Check if one of the source files changed
                changes = False
                for source in cls._sources.itervalues():
                    latestmtime = source.c4pfile.update()
                    if latestmtime and ((not source in cls._mtimes) or (latestmtime > cls._mtimes[source])):
                        cls._mtimes[source] = latestmtime
                        if source.trusted:
                            cls._revoked_licenses.update(source.c4pfile.disabled_licenses)
                        changes = True

                if changes:
                    # If a profile has changed for a given appid in one of the
                    # source files, then we need to reevaluate whether it's the
                    # profile that best matches the current product for that
                    # appid, or whether a profile from another source file
                    # should be used instead.
                    # The simplest way to do so, and possibly fastest too, is
                    # to rebuild the list from scratch.
                    profiles = C4ProfilesSet()
                    profile_lists = []
                    for source in cls._sources.itervalues():
                        try:
                            # Also refreshes that source's profile list
                            profile_list = source.c4pfile.profiles
                            if profile_list is None:
                                # If the file does not exist, then we get
                                # None with no exception
                                continue
                        # pylint: disable-msg=W0703
                        except Exception, exception:
                            cxlog.err("an error occurred while reading %s:\n%s" % (cxlog.debug_str(source), cxlog.debug_str(exception)))
                            continue

                        # First check all files for revoke profiles, so we know
                        # new revoke profiles are applied everywhere
                        if source.trusted and source.c4pfile.revokelist:
                            revokelist_file = c4profiles.get_revokelist_file()

                            revokelist_file.lock_file()
                            try:
                                for revoke_profile in source.c4pfile.revokelist:
                                    already_registered = False
                                    revoke_string = '%s..%s' % (
                                        revoke_profile.first and revoke_profile.first.encode('utf8') or '',
                                        revoke_profile.last and revoke_profile.last.encode('utf8') or '')
                                    section = revokelist_file[revoke_profile.appid.encode('utf8')]
                                    for key in section:
                                        if section[key] == revoke_string:
                                            already_registered = True
                                            break

                                    if not already_registered:
                                        section['Range%s' % len(section)] = revoke_string
                            finally:
                                revokelist_file.save_and_unlock_file()

                        profile_lists.append(profile_list)

                    for profile_list in profile_lists:
                        for profile in profile_list:
                            if profile.is_revoked:
                                continue

                            appid = profile.appid
                            if appid not in profiles:
                                profiles[appid] = profile
                            elif profile.score < profiles[appid].score:
                                profiles[appid] = profile

                    # Atomically update cls._profiles
                    cls._profiles = profiles
                    cls._last_checked = time.time()
            finally:
                cls._lock.release()
        return cls._profiles

    @classmethod
    def load_tie_files(cls):
        for basename in ("crossover.tie", "unknown.tie", "internal.tie", "unsupported.tie"):
            filename = os.path.join(cxutils.CX_ROOT, "share/crossover/data", basename)
            cls.add_c4p_file(filename, None, c4profiles.builtin)

        if os.path.exists(cls.tiedir):
            for c4pfile in os.listdir(cls.tiedir):
                if os.path.splitext(c4pfile)[1] == ".c4p" or os.path.splitext(c4pfile)[1] == ".tie":
                    if c4pfile == os.path.basename(online_file):
                        cls.add_c4p_file(os.path.join(cls.tiedir, c4pfile), None, c4profiles.download)
                    else:
                        cls.add_c4p_file(os.path.join(cls.tiedir, c4pfile), None, c4profiles.dropin)

    @classmethod
    def get_revoked_licenses(cls):
        """Returns a set of revoked license id's."""
        cls.all_profiles() # reload ties as needed
        return cls._revoked_licenses

    def copy(self):
        # We have to override this because IterableUserDict wrongly copies userdata.
        result = C4ProfilesSet()
        result.data.update(self.data)
        return result


online_file = os.path.join(C4ProfilesSet.tiedir, "crossover.tie")

@cxobjc.method(C4Profiles, 'getOnlineFilePath')
def getOnlineFilePath():
    # This is used by the Mac GUI to display the mtime of the file.
    return online_file

# Return True if anything is updated.
def update_online_profiles(url):
    try:
        os.makedirs(os.path.join(C4ProfilesSet.tiedir, "download"))
    except os.error:
        pass

    it_was_there = False
    if os.path.exists(online_file):
        it_was_there = True
        rfc822format = "%a, %d %b %Y %H:%M:%S UTC"
        mtime = datetime.datetime.utcfromtimestamp(os.path.getmtime(online_file)).strftime(rfc822format)
    else:
        mtime = ""
    tmpfile = os.path.join(C4ProfilesSet.tiedir, "download", os.path.basename(url))

    # Download the file if it is newer than the one we have.
    getter = cxurlget.UrlGetter(url, tmpfile, last_modified=mtime)
    getter.fetch()

    if getter.finished and os.path.exists(tmpfile):
        # Make sure that the downloaded file is properly signed.  If
        #  not then discard it.
        if not C4ProfilesSet.taste_new_c4p_file(tmpfile):
            cxlog.warn("The newly downloaded %s file is not properly signed. Discarding." % tmpfile)
            os.remove(tmpfile)
            return False

        [newfile, sig] = C4ProfilesSet.prepare_external_c4p(tmpfile)
        if not it_was_there:
            C4ProfilesSet.add_c4p_file(newfile, sig, c4profiles.download)
        if newfile != online_file:
            cxlog.warn("We are downloading profile file %s but expecting it to be named %s. That means that our mtime comparison is invalid." % (newfile, online_file))
        return True
    else:
        return False

