# (c) Copyright 2009-2010. CodeWeavers, Inc.

import os

import cxlog
import distversion

import urllib2
import cxobjc
import binascii
import time
import tempfile
import c4profilesmanager
import proxyinfo

import cxutils
import cxconfig

# for localization
from cxutils import cxgettext as _


class DemoUtils(cxobjc.Proxy):
    pass

@cxobjc.method(DemoUtils, 'demoStatusForLicenseFile_andSig_')
def demo_status(licenseFile, sigFile):
    """Checks the demo-status of the install.  Returns a 5-tuple.
       The first item is a simple boolean:  True if we're in demo
       mode, False if we aren't.  The second item is a string which
       will describe the account used to register the demo, if any.
       The third entry is the expiration date of the license.
       Fourth entry:  License ID
       Fifth entry:  True if this license has been revoked
    """

    keyFile = os.path.join(cxutils.CX_ROOT, "share/crossover/data", "tie.pub")

    if not (os.path.exists(licenseFile) and os.path.exists(sigFile)):
        cxlog.log("Unable to locate license %s and/or signature %s.\n" % (licenseFile, sigFile))
        return (True, None, None, None, False)

    if not signatureIsValid(licenseFile, sigFile, keyFile):
        cxlog.log("Signature %s for file %s is invalid.\n" % (sigFile, licenseFile))
        return (True, None, None, None, False)

    licenseData = cxconfig.get(licenseFile)
    username = licenseData[distversion.DEMO_SKU].get('customer')
    expiration = licenseData[distversion.DEMO_SKU].get('expires')
    licenseID = licenseData["license"].get('id')

    if not expiration or not username or not licenseID:
        # The license file is there, but there's no entry for this product.
        return (True, None, None, None, False)

    if licenseID in c4profilesmanager.C4ProfilesSet.get_revoked_licenses():
        return (True, username, expiration, licenseID, True)

    builddate = time.strptime(distversion.DATE,"%Y/%m/%d")
    expirationdate = time.strptime(expiration,"%Y/%m/%d")

    if builddate > expirationdate:
        return (True, username, expiration, licenseID, False)
    else:
        return (False, username, expiration, licenseID, False)


@cxobjc.method(DemoUtils, 'canRegisterWithLicenseDir_')
def can_register_as_current_user(licenseDir):
    """ Returns True if we have current permission to register.  Return
        False if we need to authenticate as root before registering. 
    """
    licenseFile = os.path.join(licenseDir, "license.txt")
    sigFile = os.path.join(licenseDir, "license.sig")

    if os.path.exists(licenseFile):
        if not os.access(licenseFile, os.W_OK):
            return False

    if os.path.exists(sigFile):
        if not os.access(sigFile, os.W_OK):
            return False

    if os.path.exists(licenseDir):
        return os.access(licenseDir, os.W_OK)

    parentDir =  os.path.dirname(licenseDir)
    return os.access(parentDir, os.W_OK)

@cxobjc.method(DemoUtils, 'removeRegMenu')
def remove_registration_menu():
    # Remove the 'register and unlock' menu item.
    args = [os.path.join(cxutils.CX_ROOT, "bin", "cxmenu"),
            "--crossover", "--mode", "ignore", "--uninstall",
            "--filter", "StartMenu/Register and Unlock this Demo"]
    menuRemovalResults = cxutils.run(args, stderr=cxutils.GRAB)
    if menuRemovalResults[0] or menuRemovalResults[2]:
        cxlog.log("Warning:  Failed to remove the registration menu item: %s" % menuRemovalResults[2])

    return menuRemovalResults[0]


@cxobjc.method(DemoUtils, 'licenseUrlForLogin_andPassword_')
def reg_license_url(login, password):
    # Get an md5 of the password.
    md5hasher = cxutils.md5_hasher()
    md5hasher.update(password)
    mdPassword = md5hasher.hexdigest()

    URL = "HTTPS://www.codeweavers.com/bin/register/%s;%s;lic;%s;md5" % (login, mdPassword, distversion.DEMO_EXTREP)
    return URL

@cxobjc.method(DemoUtils, 'splitLicenseAndSigFromBuffer_')
def splitLicenseAndSigFromBuffer(licenseBuffer):
    [openLicenseFD, licenseTempFile] = tempfile.mkstemp(prefix = "license.txt")
    openLicense = os.fdopen(openLicenseFD,"w")

    [openSigFD, sigTempFile] = tempfile.mkstemp(prefix = "license.sig")
    openSig = os.fdopen(openSigFD,"w")

    rVals = (licenseTempFile, sigTempFile)

    sigannounce = "<!-- Signature"
    inSig = False

    # Split the downloaded file into license and signature, and write them out.
    for l in licenseBuffer.split(os.linesep):
        if sigannounce in l:
            inSig = True

        if inSig:
            # We have a signature embedded in an XML comment.
            # Strip out the commenty bits.
            strippedline = l.replace(sigannounce,"")
            strippedline = strippedline.replace("-->","")
            strippedline = strippedline.rstrip()
            splitline = strippedline.split("= ")
            if len(splitline) > 1:
                hexline = splitline[1].strip()
            else:
                hexline = strippedline.strip()

            openSig.write(binascii.unhexlify(hexline))
        else:
            openLicense.write(l)
            openLicense.write("\n")

    openLicense.close()
    openSig.close()

    if not inSig:
        cxlog.log("The license file we downloaded has no signature, so is probably broken.")
        os.remove(rVals[0])
        os.remove(rVals[1])

        return (None, None)
    else:
        return rVals

def download_and_split_license_file(login, password):
    """ Attempt to download a license file and drop it into the appropriate dir.
        Returns licensefile, sigfile, errorstring where filename will be None 
        if the download failed.  Otherwise licensefile and sigfile are temporary files;
        it is the caller's responsibility to test, move, and/or delete them.
    """

    errorString = None
    licenseFile = None
    sigFile = None

    URL = reg_license_url(login, password)
    lastResort = _("If this problem continues, contact CodeWeavers sales support at sales@CodeWeavers.com.")

    while True:

        try:
            proxyinfo.install_default_opener()            
            licenseFileLike = urllib2.urlopen(URL)
            licenseBuffer = licenseFileLike.read()
        except urllib2.URLError:
            errorString = _("Download failed.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break

        if licenseBuffer.find("FILE NOT FOUND", 0, 256) != -1:
            errorString = _("Unable to locate an upgrader for %(product)s.\n\n%(lastresort)s") % {'product': distversion.PRODUCT_NAME, 'lastresort': lastResort}
            break
        elif licenseBuffer.find("USER NOT FOUND", 0, 256) != -1 or licenseBuffer.find("BAD PASSWORD", 0, 256) != -1:
            errorString = _("Unable to validate your login and password.\n\nVisit store.codeweavers.com to make sure you have a valid login, and try again.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break
        elif licenseBuffer.find("ERROR", 0, 256) != -1:
            errorString = _("Download failed.\n\n%(lastresort)s") % {'lastresort': lastResort}
            break

        (licenseFile, sigFile) = splitLicenseAndSigFromBuffer(licenseBuffer)

        if not licenseFile or not sigFile:
            errorString = _("Download corrupt.\n\n%(lastresort)s") % {'lastresort': lastResort}
            licenseFile = None
            sigFile = None
            break

        break

    return licenseFile, sigFile, errorString


def signatureIsValid(fileName, sigFileName, keyFileName):
    args = ["openssl", "dgst", "-sha1", "-verify", keyFileName, "-signature", sigFileName, fileName]
    verification_result = cxutils.run(args, stdout=cxutils.NULL, stderr=cxutils.NULL)
    if verification_result[0] == 0:
        return True
    else:
        return False

